; ==============================================================================
; file: "source/boot.asm"
; use: to be compiled into "bootsect"
; ==============================================================================

use16                                   ; bios executes bootloader in real mode
org 0x7C00

load_kernel:

    .setup_segments_and_stack_pointer:
        mov bx, 0                       ; bios loads bootsector at 0x00007C00
        mov ds, bx                      ; and we set org directive to 0x7C00
        mov es, bx                      ; so segments equal zero
        mov ss, bx
        mov sp, 0xFFF0                  ; stack grows downwards from 0x0000FFF0

    .save_device:
        mov byte [bootdev], dl          ; we save boot device

    .check_bios_lba_extensions:
        mov ah, 0x41                    ; we check if bios lba extensions exist
        mov bx, 0x55AA
        int 0x13
        jc fail                         ; carry flag is set if they don't exist
        cmp bx, 0xAA55                  ; and bx is 0xAA55 if they exist
        jne fail

    .read_rest_of_bootloader_and_superblock:
        mov eax, 1                      ; rest of bootloader resides on sector 1
        mov bx, 0                       ; and is loaded to 0x00007E00
        mov fs, bx
        mov bx, 0x7E00
        mov cx, 3                       ; superblock follows and is also loaded,
        call read_sectors               ; so 3 sectors are loaded using bios
    
    .check_ext2:
        cmp word [signature], 0xEF53    ; ext2 magic signature is 0xEF53
        jne fail

    .process_block_size:
        mov cl, byte [block_size]       ; block of size 1024 has 2 sectors
        mov dword [block_size], 2       ; so 2 is shifted left by the superblock
        shl dword [block_size], cl      ; field corresponding to block size
    
    .check_block_size:
        cmp dword [block_size], 2       ; if block size in sectors is 2
        je .read_sector_4               ; group descriptor starts in block 2
    
    .read_block_1:
        mov eax, dword [block_size]     ; else it starts in block 1
        jmp .read_group_descriptor
        
    .read_sector_4:
        mov eax, 4                      ; load block 2
    
    .read_group_descriptor:
        mov bx, 0x1000                ; group descriptor is loaded at 0x00001000
        mov cx, 8                       ; temporarily 8 sectors are loaded
        call read_sectors

    .read_inode_2:
        mov word [ram_offset], 0x4000   ; inode 2 is the root directory inode
        mov word [ram_segment], 0       ; and here it is loaded at 0x00004000
        mov dword [inode], 2
        call read_inode
    
    .read_system_dir:
        mov word [ram_offset], 0x4000   ; current inode is loaded at 0x00004000
        mov si, dir_name                ; and it contains the system directory
        call read_entry                 ; if it is found, else system hangs
    
    .read_kernel:
        mov word [ram_offset], 0        ; current inode is loaded at 0x00020000
        mov word [ram_segment], 0x2000  ; and it contains the kernel file
        mov si, kernel_name             ; if it is found, else system hangs
        call read_entry
        jmp change_video_mode           ; loading finished for now

; ------------------------------------------------------------------------------

read_sectors:
    
    save_registers:
        pusha                           ; registers are preserved
    
    .build_disk_address_packet:
        mov si, 7BE0h                   ; disk address packet is at 0x00007BE0
        mov word [ds:si], 0x10          ; size of packet is 16 bytes
        mov word [ds:si + 2], cx        ; sector count is in cx
        mov word [ds:si + 4], bx        ; memory offset is in bx
        mov word [ds:si + 6], fs        ; memory segment is in fs
        mov dword [ds:si + 8], eax      ; low starting lba sector is in eax
        mov dword [ds:si + 12], 0       ; high starting lba sector is 0,
                                        ; who needs more than 2 terabytes?
    .invoke_int_13h:
        mov ah, 0x42                    ; bios extended read is used here
        mov dl, byte [bootdev]          ; boot device will be read
        int 0x13                        ; bios disk interrupt is invoked
    
    .return:
        popa                            ; registers are restored
        ret                             

; ------------------------------------------------------------------------------

compare_string:

        pusha                           ; registers are preserved

    .check_next:
        mov al, byte [si]               ; byte of first string is compared
        cmp al, byte [di]               ; with byte of second string
        jne .return                     ; if not equal, failure
        inc si                          ; else, we increment to next byte
        inc di
        dec cl                          ; we decrement remaining bytes
        cmp cl, 0                       ; if 0, equality flag is set
        jne .check_next                 ; else, next byte is checked
    
    .return:
        popa                            ; registers are restored
        ret

; ------------------------------------------------------------------------------

read_entry:
    
    .save_registers:
        pusha                           ; registers are preserved
    
    .get_previous_dir_offset:
        mov bx, 0x3000                  ; prev entry was loaded at 0x00003000
    
    .check_entry:
        mov di, bx                      ; filename starts at offset 6 of entry
        add di, 6
        mov cl, byte [si]               ; we will compare filename length,
        add cl, 2                       ; file type and filename, all together
        call compare_string             ; equal filename length + 2
        je .load                        ; if strings equal, we want this entry
    
    .go_to_next:
        movzx ax, byte [bx + 6]         ; else, we need to go to the next entry
        mov dx, 0                       ; if filename length equals
        mov cx, 4                       ; a multiple of 4, there is no padding
        idiv cx                         ; and the next entry starts immediately
        cmp dx, 0                       ; else, there is some padding
        je .add_no_pad                  ; to match the next multiple of 4
    
    .add_pad:
        shl ax, 2
        add bx, ax                      ; name without four last bytes
        add bx, 12                      ; before name + four last bytes
        jmp .check_entry                ; including padding
    
    .add_no_pad:
        shl ax, 2
        add bx, ax                      ; full name
        add bx, 8                       ; before name
        jmp .check_entry
        
    .load:
        mov eax, dword [bx]             ; inode number is in the first dword
        mov dword [inode], eax          ; of the wanted entry
        call read_inode                 ; we simply load that inode
        popa                            ; registers are restored
        ret

; ------------------------------------------------------------------------------

fail:
    jmp $                               ; failure, infinite loop

; ------------------------------------------------------------------------------

strings:
    dir_name db 6,2,"system"            ; 6byte filename, directory, system
    kernel_name db 10,1,"kernel.bin"    ; 10byte filename, regular file, kernel

; ------------------------------------------------------------------------------

end_boot_sector:

    times 510 - ($ - $$) db 0           ; the rest of first sector is padded
    dw 0xAA55                           ; with zeros until the boot signature
                                        ; based on which the bios decides
                                        ; to load or not the boot sector
                                        ; in other words NEVER REMOVE IT!!!

; ==============================================================================
; sector 1 starts here
; ==============================================================================

read_inode:
        
        pusha                           ; registers are preserved
        dec dword [inode]               ; inode calculations are easier this way
        mov word [.count], 0            ; block counter is 0

    .get_group_number:
        mov eax, dword [inode]
        mov edx, 0
        idiv dword [inodes_per_group]
        mov dword [group], eax          ; group = (inode - 1) / inodes_per_group
        mov dword [index], edx          ; index = (inode - 1) % inodes_per_group
    
    .read_inode_table:
        shl ax, 5                       ; each group descriptor occupies
        mov bx, ax                      ; 32 bytes and starts at 0x1000
        mov eax, dword [bx+0x1000+8]    ; third dword contains the starting
        imul eax, dword [block_size]    ; block of inode table which
        mov bx, 0                       ; is loaded at 0x00002000
        mov fs, bx
        mov bx, 0x2000
        mov cx, 8                       ; temporarily 8 sectors are loaded
        call read_sectors
    
    .read_direct_block:
        mov bx, word [inode_struct_size] ; struct_size * index = offset_in_table
        imul bx, word [index]
        add bx, word [.count]           ; offset 40 (first direct block)
        add bx, word [.count]           ; added with (direct_block_counter * 4)
        add bx, word [.count]           ; equals offset of wanted direct block
        add bx, word [.count]
        mov eax, dword [bx+0x2000+40]   ; so we take this block number
        cmp eax, 0                      ; if 0, pointer to bootloader, invalid
        je .return
        call read_direct                ; else, we read that direct block
    
    .increment_counter:
        inc word [.count]
        cmp word [.count], 12           ; there are 12 direct blocks
        jb .read_direct_block
    
    .read_indirect_block:
        mov bx, word [inode_struct_size] ; struct_size * index = offset_in_table
        imul bx, word [index]
        mov eax, dword [bx+0x2000+88]   ; indirect block pointer is at offset 88
        cmp eax, 0                      ; if 0, pointer to bootloader, invalid
        je .return
        call read_indirect              ; else, we read that indirect block

    .return:
        inc dword [inode]               ; inode number is restored
        popa                            ; registers are restored
        ret
    
    .count dw 0

; ------------------------------------------------------------------------------

read_direct:
    
    .save_registers:
        pusha                           ; registers are preserved
    
    .read_sectors:
        imul eax, dword [block_size]    ; lba_sector = block_number * block_size
        mov bx, word [ram_offset]       ; we load direct block at input address
        mov fs, word [ram_segment]
        mov cx, word [block_size]       ; one block is loaded and so are loaded
        call read_sectors               ; that many sectors as exist in it
    
    .increment_address:
        mov ax, word [block_size]       ; next block will be loaded further so
        shl ax, 5                       ; we add block size in bytes to offset
        add word [ram_segment], ax
    
    .return:
        popa                            ; registers are restored
        ret

; ------------------------------------------------------------------------------

read_indirect:
    
    .save_registers:
        pusha                           ; registers are preserved
    
    .read_sectors:
        imul eax, dword [block_size]    ; lba_sector = block_number * block_size
        mov bx, 0                       ; we load indirect block at 0x00003000
        mov fs, bx
        mov bx, 0x3000
        mov cx, word [block_size]       ; one block is loaded and so are loaded
        call read_sectors               ; that many sectors as exist in it
        mov word [.count], 0            ; block counter is 0
    
    .read_direct_block:
        mov bx, word [.count]           ; direct block pointers start at
        shl bx, 2                       ; 0x00003000 and have a dword size
        mov eax, dword [bx+0x3000]
        cmp eax, 0                      ; if 0, pointer to bootloader, invalid
        je .return
        call read_direct                ; indirect pointer points to direct ones
                                        ; so we load a direct pointer now
    .calculate_pointers_in_block:
        mov ax, word [block_size]       ; in each block there are
        shl ax, 7                       ; (block_size_in_sectors * 128) pointers
        
    .increment_counter:
        inc word [.count]               ; we go to the next block pointer
        cmp word [.count], ax           ; if block finished, no more pointers
        jb .read_direct_block
    
    .return:
        popa                            ; registers are restored
        ret
    
    .count dw 0

; ------------------------------------------------------------------------------

change_video_mode:

    .set_640x480_truecolor:
        mov ax, 0x4F02                  ; using vesa bios extensions we set
        mov bx, 0x4112                  ; graphical 640x480x24 or 640x480x32
        int 0x10
        
    .get_mode_info:
        mov ax, 0x4F01                  ; we get the info table about this mode
        mov cx, 0x4112                  ; and save it to 0x00007800
        mov di, 0x7800                  ; it is needed by display functions
        int 0x10

; ------------------------------------------------------------------------------

get_ehci_bus_dev_fun:

; TODO for 0.2.0: find ehci using pci, we must be bios independable
   
    .get_ehci_address:
        mov si, 0                       ; get pci address of first ehci ctrller
        mov ax, 0xB103
        mov ecx, 0x000C0320
        int 0x1A
    
    .check_ehci:
        cmp ah, 0x86                    ; if ah returned 0x86, we failed
        je fail
    
    .copy_bus:
        movzx eax, bh                   ; bus is in bh
    
    .copy_dev:
        movzx ecx, bl                   ; dev is in 5 higher bits of bl
        shr ecx, 3
    
    .copy_fun:
        movzx edx, bl                   ; fun is in 3 lower bits of bl
        and edx, 7

    .set_var_segment:
        mov si, 0x1200                  ; variables are saved
        mov ds, si                      ; starting at 0x00012000
        
    .save_vars:
        mov si, 0
        mov dword [ds:si], eax
        mov dword [ds:si + 4], ecx
        mov dword [ds:si + 8], edx

; ------------------------------------------------------------------------------

set_protected_mode:

    .enable_A20:
        mov ax, 0x2401                  ; A20 line is enabled using the bios
        int 0x15
        
    .load_gdt:
        cli                             ; interrupts are cleared
        mov ax, 0                       ; ds equals now 0
        mov ds, ax
        lgdt [gdt_info]                 ; global descriptor table is loaded
        
    .set_32_bits:
        mov eax, cr0                    ; cr0 lowest bit is set
        or eax, 1                       ; so protected mode is enabled
        mov cr0, eax
        jmp 0x08:escape_bootloader      ; jump to 32 bit code

; ==============================================================================
; 32 bit code starts here
; ==============================================================================

use32

escape_bootloader:

    .set_segment:
        mov ax, 0x10                    ; segments are aligned to data selector
        mov ds, ax
        mov es, ax
        mov fs, ax
        mov gs, ax
        mov ss, ax
    
    .set_stack:
        mov esp, 0x00010000             ; stack grows downwards from 0x00010000
        
    .jump:
        jmp 0x00020000                  ; kernel was loaded at 0x00020000

; ------------------------------------------------------------------------------

start_gdt:
    
    .null:
        dd 0                            ; null table has only 0
        dd 0
    .code:
        dw 0xFFFF                       ; limit bits 12-27
        dw 0                            ; base bits 0-15
        db 0                            ; base bits 16-23
        db 0b10011010                   ; access
        db 0b11001111                   ; granularity + limit bits 28-31
        db 0                            ; base bits 24-31
    .data:
        dw 0xFFFF                       ; limit bits 12-27
        dw 0                            ; base bits 0-15
        db 0                            ; base bits 16-23
        db 0b10010010                   ; access
        db 0b11001111                   ; granularity + limit bits 28-31
        db 0                            ; base bits 24-31
            
gdt_info:

    dw gdt_info - start_gdt - 1         ; length of gdt - 1
    dd start_gdt                        ; start position of gdt

; ------------------------------------------------------------------------------

variables:

    ; bootloader variables
    bootdev           equ 0x7A00
    inode             equ 0x7A04
    group             equ 0x7A08
    index             equ 0x7A0C
    ram_offset        equ 0x7A10
    ram_segment       equ 0x7A14

    ; superblock variables
    total_inodes      equ 0x8000 + 0
    total_blocks      equ 0x8000 + 4
    block_size        equ 0x8000 + 24
    blocks_per_group  equ 0x8000 + 32
    inodes_per_group  equ 0x8000 + 40
    signature         equ 0x8000 + 56
    inode_struct_size equ 0x8000 + 88

end_bootloader:

    times 1024 - ($ - $$) db 0         ; rest of bootloader is padded with zeros
