; ==============================================================================
; file: "souce/system/kernel/ext2.inc"
; use: to be included in "source/system/kernel.asm"
; ==============================================================================

ext2_read_file:

; ------------------------------------------------------------
; description: reads the file of the given path
; input: dword [string_1] = address of path string
;        dword [ram_address] = memory address to load the file
; output: {none}
; altered: dword [lba_sector]
;          dword [sector_count]
;          dword [ext2_inode]
;          dword [ext2_group]
;          dword [ext2_index]
; ------------------------------------------------------------

    .prepare:
        pusha                           ; registers are preserved
        push dword [ram_address]        ; save address

    .read_superblock:
        mov dword [ram_address], 0x00060000 ; superblock is read at 0x00060000
        mov dword [lba_sector], 2           ; it starts at sector 2
        mov dword [sector_count], 2         ; it contains 2 sectors
        call ehci_read_sectors
    
    .check_ext2:
        cmp word [ext2_signature], 0xEF53   ; ext2 magic signature is 0xEF53
        jne .return
    
    .process_block_size:
        mov cl, byte [ext2_block_size]  ; block of size 1024 has 2 sectors
        mov dword [ext2_block_size], 2  ; so 2 is shifted left by the superblock
        shl dword [ext2_block_size], cl ; field corresponding to block size
    
    .read_group_descriptor:
        cmp dword [ext2_block_size], 2  ; if block size in sectors is 2
        je .read_sector_4               ; group descriptor starts in block 2
    
    .read_block_1:
        mov eax, dword [ext2_block_size] ; else it starts in block 1
        jmp .continue
    
    .read_sector_4:
        mov eax, 4                       ; load block 2
    
    .continue:
        mov dword [lba_sector], eax
        mov dword [sector_count], 8         ; temporarily 8 sectors are fetched
        mov dword [ram_address], 0x00064000 ; group descriptor is loaded
        call ehci_read_sectors              ; at 0x00064000
    
    .read_inode_2:
        mov dword [ram_address], 0x0007C000 ; root directory is inode 2
        mov dword [ext2_inode], 2           ; and is loaded at 0x0007C000
        call ext2_read_inode
        
    .analyze_path:
        mov esi, dword [string_1]       ; we skip "/dev0"
        add esi, 5
    
    .copy_entry:
        add esi, 1                      ; we go to the next char
        mov edi, ext2_entry+2           ; we start copying the filename
        mov ecx, 0                      ; to the filename buffer + 2
    
    .copy_byte:
        mov al, byte [esi]              ; if byte in esi is "/"
        cmp al, "/"                     ; we need to find a dir
        je .set_dir
        cmp al, 0                       ; if it is null termination
        je .set_file                    ; we need to find a regular file
        mov byte [edi], al            ; else, not end of filename and we save it
        add esi, 1                      ; we go to next char
        add edi, 1
        add ecx, 1                      ; length of filename until now is
        jmp .copy_byte               ; incremented and we have to copy this byte
    
    .set_dir:
        mov edi, ext2_entry
        mov byte [edi], cl              ; at filename buffer there is length
        mov byte [edi+1], 2             ; at filename buffer + 1 there is type
                                        ; directory type is 0x02
    .read_dir:
        push esi                        ; full file path pointer is preserved
        mov esi, edi                    ; esi points to filename buffer
        call ext2_read_entry            ; and that file is loaded
        pop esi                         ; full file path pointer is restored
        jmp .copy_entry
    
    .set_file:
        mov edi, ext2_entry
        mov byte [edi], cl              ; at reserved address there is length
        mov byte [edi+1], 1             ; at reserved address + 1 there is type
                                        ; regular file type is 0x01
    .read_file:
        pop dword [ram_address]         ; restore address
        push esi                        ; full file path pointer is preserved
        mov esi, edi                    ; esi points to filename buffer
        call ext2_read_entry            ; and that file is loaded
        pop esi                         ; full file path pointer is restored

    .return:
        popa                            ; registers are restored
        ret

; ==============================================================================

ext2_read_entry:

; --------------------------------------------------------------------------
; description: scans a directory loaded at 0x0006C000 for a given entry
; input: esi = address of filename to search
;              (esi = name length, esi+1 = file type, esi+2 = start of name)
; output: {none}
; altered: dword [ext2_inode]
;          dword [ext2_group]
;          dword [ext2_index]
;          dword [lba_sector]
;          dword [sector_count]
; --------------------------------------------------------------------------

    .prepare:
        pusha                           ; registers are preserved
        mov ebx, 0x0007C000             ; prev entry was loaded at 0x0006C000
        
    .check_entry:
        mov edi, ebx                    ; we need to point edi to the entry
        add edi, 6                      ; so we can compare filename given
        mov cl, byte [esi]              ; and filename of the entry
        add cl, 2
        call ext2_compare_filename
        je .read                        ; equality flag is set if they equal
    
    .go_to_next:
        movzx eax, byte [ebx+6]         ; else we need to go to the next entry
        mov edx, 0                      ; if filename length equals
        mov ecx, 4                      ; a multiple of 4, there is no padding
        div ecx                         ; and the next entry starts immediately
        cmp edx, 0                      ; else, there is some padding
        je .add_no_pad                  ; to match the next multiple of 4
    
    .add_pad:
        shl eax, 2                      ; if filename length not a multiple of 4
        add ebx, eax                    ; name without four last bytes
        add ebx, 12                     ; before name + four last bytes
        jmp .check_entry                ; including padding
    
    .add_no_pad:
        shl eax, 2
        add ebx, eax                    ; full name
        add ebx, 8                      ; before name
        jmp .check_entry
    
    .read:
        mov eax, dword [ebx]            ; inode number is the first dword
        mov dword [ext2_inode], eax     ; of the wanted entry
        call ext2_read_inode            ; we simply load that inode
    
    .return:
        popa                            ; registers are restored
        ret

; ==============================================================================

ext2_compare_filename:

; --------------------------------------------------------------------------
; description: compares if the provided filename mathes this one found
; input: esi = address of first filename
;              (esi = name length, esi+1 = file type, esi+2 = start of name)
;        edi = address of second filename
;              (edi = name length, edi+1 = file type, edi+2 = start of name)
;        cl = length of name + 2
; output: equality flag (set if equal, clear if differ)
; altered: {none}
; --------------------------------------------------------------------------

    .save_registers:
        pusha                           ; registers are preserved
    
    .check_next:
        mov al, byte [esi]              ; byte of first filename is compared
        cmp al, byte [edi]              ; with byte of second filename
        jne .return                     ; if not equal, failure
        inc esi                         ; else, we increment to next byte
        inc edi
        dec cl                          ; we decrement remaining bytes
        cmp cl, 0                       ; if 0, equality flag is set
        jne .check_next                 ; else, next byte is checked
   
    .return:
        popa                            ; registers are restored
        ret

; ==============================================================================
    
ext2_read_inode:

; -------------------------------------------------------------
; description: reads an inode
; input: dword [ext2_inode] = number of inode
;        dword [ram_address] = memory address to load the inode
; output: {none}
; altered: dword [ext2_group]
;          dword [ext2_index]
;          dword [lba_sector]
;          dword [sector_count]
; -------------------------------------------------------------
        
    .prepare:
        pusha                           ; registers are preserved
        dec dword [ext2_inode]          ; inode calculations are easier this way
        mov dword [.count], 0           ; block counter is 0
    
    .get_group_number:
        mov eax, dword [ext2_inode]
        mov edx, 0
        div dword [ext2_inodes_per_group]
        mov dword [ext2_group], eax     ; group = (inode - 1) / inodes_per_group
        mov dword [ext2_index], edx     ; index = (inode - 1) % inodes_per_group
    
    .read_inode_table:
        shl eax, 5                        ; each group descriptor occupies
        mov ebx, dword [eax+0x00064000+8] ; 32 bytes and starts at 0x00064000
        imul ebx, dword [ext2_block_size] ; third dword contains the starting
        mov dword [lba_sector], ebx       ; block of inode table
        mov dword [sector_count], 8       ; temporarily 8 sectors are loaded
        push dword [ram_address]
        mov dword [ram_address], 0x00068000 ; inode table is read at 0x00068000
        call ehci_read_sectors
        pop dword [ram_address]           
    
    .read_direct_block:
        mov ebx, dword [ext2_inode_struct_size] ; struct_size * index
        imul ebx, dword [ext2_index]            ; equals offset_in_table
        add ebx, dword [.count]                 ; offset 40 (first direct block)
        add ebx, dword [.count]         ; added with (direct_block_counter * 4)
        add ebx, dword [.count]         ; equals offset of wanted direct block
        add ebx, dword [.count]
        mov eax, dword [ebx+0x00068000+40]      ; so we take this block number
        cmp eax, 0                      ; if 0, pointer to bootloader, invalid
        je .return
        call ext2_read_direct           ; else, we read that direct block
    
    .increment_counter:
        inc dword [.count]
        cmp dword [.count], 12          ; there are 12 direct blocks
        jb .read_direct_block
        
    .read_indirect_block:
        mov ebx, dword [ext2_inode_struct_size] ; struct_size * index
        imul ebx, dword [ext2_index]            ; equals offset_in_table
        mov eax, dword [ebx+0x00068000+88]      ; indirect block pointer
        cmp eax, 0                              ; is at offset 88
        je .return                      ; if 0, pointer to bootloader, invalid
        call ext2_read_indirect         ; else, we read that indirect block
    
   .read_doubly_indirect_block:
        mov ebx, dword [ext2_inode_struct_size] ; struct_size * index
        imul ebx, dword [ext2_index]            ; equals offset_in_table
        mov eax, dword [ebx+0x00068000+92]      ; doubly indirect block pointer
        cmp eax, 0                              ; is at offset 92
        je .return                      ; if 0, pointer to bootloader, invalid
        call ext2_read_doubly_indirect  ; else, we read that doubly indirect
    
    .return:
        inc dword [ext2_inode]          ; inode number is restored
        popa                            ; registers are restored
        ret
        
    .count dd 0

; ==============================================================================

ext2_read_direct:

; -------------------------------------------------------------
; description: reads direct block
; input: eax = block number to read
;        dword [ram_address] = memory address to load the block
; output: dword [ram_address] = where to load the next block
; altered: dword [lba_sector]
;          dword [sector_count]
; -------------------------------------------------------------
    
    .save_registers:
        pusha                           ; registers are preserved
    
    .calculate_sector:
        imul eax, dword [ext2_block_size] ; lba_sector = block_number*block_size
        mov dword [lba_sector], eax
    
    .calculate_sector_count:
        mov eax, dword [ext2_block_size] ; one block is loaded and so are
        mov dword [sector_count], eax    ; that many sectors as exist in it
    
    .read_sectors:
        call ehci_read_sectors
    
    .increment_address:
        mov eax, dword [ext2_block_size] ; next block will be loaded further so
        shl eax, 9                       ; we add block size in bytes to offset
        add dword [ram_address], eax
    
    .return:
        popa                            ; registers are restored
        ret

; ==============================================================================

ext2_read_indirect:

; -------------------------------------------------------------
; description: reads indirect block
; input: eax = block number to read
;        dword [ram_address] = memory address to load the block
; output: dword [ram_address] = where to load the next block
; altered: dword [lba_sector]
;          dword [sector_count]
; -------------------------------------------------------------
    
    .save_registers:
        pusha                           ; registers are preserved
    
    .calculate_sector:
        imul eax, dword [ext2_block_size] ; lba_sector = block_number*block_size
        mov dword [lba_sector], eax
    
    .calculate_sector_count:
        mov eax, dword [ext2_block_size] ; one block is loaded and so are
        mov dword [sector_count], eax    ; that many sectors as exist in it
    
    .read_sectors:
        push dword [ram_address]        ; indirect block is loaded at 0x0006C000
        mov dword [ram_address], 0x0006C000
        call ehci_read_sectors
        pop dword [ram_address]
        mov dword [.count], 0           ; block counter is reset
    
    .read_direct_block:
        mov ebx, dword [.count]           ; direct block pointers start at
        mov eax, dword [ebx*4+0x0006C000] ; 0x0006C000 and have a dword size
        cmp eax, 0                      ; if 0, pointer to bootloader, invalid
        je .return
        call ext2_read_direct           ; indirect pointer points to direct ones
                                        ; so we load a direct pointer now
    .calculate_pointers_in_block:
        mov eax, dword [ext2_block_size] ; in each block there are
        shl eax, 7                      ; (block_size_in_sectors * 128) pointers
    
    .increment_counter:
        add dword [.count], 1           ; we go to the next block pointer
        cmp dword [.count], eax         ; if block finished, no more pointers
        jb .read_direct_block
    
    .return:
        popa                            ; registers are restored
        ret
    
    .count dd 0

; ==============================================================================

ext2_read_doubly_indirect:

; -------------------------------------------------------------
; description: reads doubly direct block
; input: eax = block number to read
;        dword [ram_address] = memory address to load the block
; output: dword [ram_address] = where to load the next block
; altered: dword [lba_sector]
;          dword [sector_count]
; -------------------------------------------------------------
    
    .save_registers:
        pusha                           ; registers are preserved
    
    .calculate_sector:
        imul eax, dword [ext2_block_size] ; lba_sector = block_number*block_size
        mov dword [lba_sector], eax
    
    .calculate_sector_count:
        mov eax, dword [ext2_block_size] ; one block is loaded and so are
        mov dword [sector_count], eax    ; that many sectors as exist in it
    
    .read_sectors:
        push dword [ram_address]            ; doubly indirect block is loaded
        mov dword [ram_address], 0x00070000 ; at 0x00070000
        call ehci_read_sectors
        pop dword [ram_address]
        mov dword [.count], 0           ; block counter is reset
    
    .read_indirect_block:
        mov ebx, dword [.count]        ; singly indirect block pointers start at
        mov eax, dword [ebx*4+0x00070000] ; 0x00070000 and have a dword size
        cmp eax, 0                      ; if 0, pointer to bootloader, invalid
        je .return
        call ext2_read_indirect         ; doubly indirect pointer
                                        ; points to singly indirect ones
                                      ; so we load a singly indirect pointer now
    .calculate_pointers_in_block:
        mov eax, dword [ext2_block_size] ; in each block there are
        shl eax, 7                      ; (block_size_in_sectors * 128) pointers
    
    .increment_counter:
        add dword [.count], 1           ; we go to the next block pointer
        cmp dword [.count], eax         ; if block finished, no more pointers
        jb .read_indirect_block
    
    .return:
        popa                            ; registers are restored
        ret
    
    .count dd 0

; ==============================================================================

ext2_variables:

; -------------------------------------------------------------------
; it is dangerous to have these variables accessible to user programs
; -------------------------------------------------------------------

    ; superblock is loaded at 0x00060000
    ext2_total_inodes      equ 0x00060000+0
    ext2_total_blocks      equ 0x00060000+4
    ext2_block_size        equ 0x00060000+24
    ext2_blocks_per_group  equ 0x00060000+32
    ext2_inodes_per_group  equ 0x00060000+40
    ext2_signature         equ 0x00060000+56
    ext2_inode_struct_size equ 0x00060000+88

    ; ext2 variables start at 0x00014000
    ext2_inode             equ 0x00014000
    ext2_group             equ 0x00014004
    ext2_index             equ 0x00014008
    ext2_entry             equ 0x00014010
